// SPDX-License-Identifier: GPL-2.0
// Copyright (c) 2008, 2017-2019 Rupert Eibauer. All rights reserved.

#include <linux/timer.h>
#include <linux/printk.h>
#include <linux/init.h>

#include <asm/system.h>
#include <asm/io.h>

//#define NO_WAIT
#ifdef NO_WAIT

#define wait() do { } while(0)

#else /* NO_WAIT */
static void wait(void) {
	volatile int wait_counter;
	volatile int wait_counter2;
	for (wait_counter = 0; wait_counter < 1000; wait_counter++)
	for (wait_counter2 = 0; wait_counter2 < 1000; wait_counter2++);
}
#endif /* NO_WAIT */

/* Simple completion test */
/* The result should look like the following (when watched with simulavr-disp)
   PORTB alternating between 1 and 2
*/

struct completion s;
struct task_struct tw;
void fw(void *arg)
{
	for (;;)
	{
		out(_PORTB,  1);
		wait();
		wait_for_completion(&s);
		printf("a\n");
	}
}

static struct timer test_timer;
static void timer_callback(void) {
	out(_PORTB, 2);
	complete_from_irq(&s);
	printf("Timer\n");
#ifdef CONFIG_DEL_TIMER
	mod_timer(&test_timer, test_timer.expires + HZ);
#else
	test_timer.expires += HZ;
	add_timer(&test_timer);
#endif
	
}
static struct timer test_timer = {
	.function = timer_callback,
	.expires  = HZ + 1 + HZ,
};

static int init_semaphore_timer_test(void) {
	int i;
	printk(KERN_NOTICE "HELLO WORLD!\n");
	out(_DDRB, 0x3f);
	out(_DDRC, 0x3f);
	local_irq_enable();
	add_timer(&test_timer);
	wait();
	wait();
	wait();
	wait();
	out(_PORTB, 7);
	for (i = 0; i <= 7; i++) {
		wait();
		out(_PORTB, i);
	}
	wait();
	out(_PORTB, 0);

	kernel_thread(&tw, fw, NULL, 2); /* High-prio waiter */
	out(_PORTB, 1);
	printk(KERN_NOTICE "calling schedule!\n");
	return 0;
}

initcall(init_semaphore_timer_test);
